import numpy as np

from occamypy import Operator
from ..vector import VectorNumpy


class FFT(Operator):
    """
    N-dimensional Fast Fourier Transform for complex input.
    
    :param model : :obj:`Vector`
        Domain vector
    :param axes : :obj:`tuple`, optional
        Dimension along which FFT is computed (all by default)
    :param nfft : :obj:`tuple`, optional
        Number of samples in Fourier Transform for each direction (same as model by default)
    :param sampling : :obj:`tuple`, optional
        Sampling steps on each axis (1. by default)
    """
    
    def __init__(self, model, axes=None, nfft=None, sampling=None):
        
        if axes is None:
            axes = tuple(range(model.ndim))
        elif not isinstance(axes, tuple) and model.ndim == 1:
            axes = (axes,)
        if nfft is None:
            nfft = model.shape
        elif not isinstance(nfft, tuple) and model.ndim == 1:
            nfft = (nfft,)
        if sampling is None:
            sampling = tuple([1.] * model.ndim)
        elif not isinstance(sampling, tuple) and model.ndim == 1:
            sampling = (sampling,)
        
        if len(axes) != len(nfft) != len(sampling):
            raise ValueError('axes, nffts, and sampling must have same number of elements')

        self.axes = axes
        self.nfft = nfft
        self.sampling = sampling
        
        self.fs = [np.fft.fftfreq(n, d=s) for n, s in zip(nfft, sampling)]
        
        dims_fft = np.asarray(model.shape)
        for a, n in zip(self.axes, self.nfft):
            dims_fft[a] = n
        
        super(FFT, self).__init__(domain=VectorNumpy(np.zeros(model.shape, dtype=np.complex)),
                                  range=VectorNumpy(np.zeros(shape=dims_fft, dtype=np.complex)))
    
    def __str__(self):
        return 'numpyFFT'
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        modelNd = model.getNdArray()
        dataNd = data.getNdArray()
        dataNd += np.fft.fftn(modelNd, s=self.nfft, axes=self.axes, norm='ortho')
        return
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        modelNd = model.getNdArray()
        dataNd = data.getNdArray()
        # here we need to separate the computation and use np.take for handling nfft > model.shape
        temp = np.fft.ifftn(dataNd, s=self.nfft, axes=self.axes, norm='ortho')
        for a in self.axes:
            temp = np.take(temp, range(self.domain.shape[a]), axis=a)
        modelNd += temp
        return


if __name__ == '__main__':
    
    import matplotlib.pyplot as plt
    
    DIM = 1
    
    if DIM == 1:
        # 1D
        dt = 0.005
        nt = 100
        t = VectorNumpy(np.arange(nt) * dt)
        f0 = 10
        nfft = 2 ** 10
        x = VectorNumpy(np.sin(2 * np.pi * f0 * t.getNdArray()))
        
        F = FFT(x, nfft=nfft, sampling=dt)
        F.dotTest(True)
        
        spectrum = F * x
        x_inv = F.H * spectrum
        
        fig, axs = plt.subplots(1, 2, figsize=(10, 4))
        axs[0].plot(t.getNdArray(), x.getNdArray(), 'k', lw=2, label='True')
        axs[0].plot(t.getNdArray(), x_inv.getNdArray().real, '--r', lw=2, label='Inverted')
        axs[0].legend()
        axs[0].set_title('Signal')
        
        axs[1].plot(F.fs[0][:F.nfft[0] // 2], np.abs(spectrum.getNdArray()[:F.nfft[0] // 2]), 'k', lw=2)
        axs[1].set_title('Fourier Transform')
        axs[1].set_xlim(0, 3 * f0)
        plt.show()
    
    elif DIM == 3:
        # 3D
        dt, dx, dy = 0.005, 5, 3
        nt, nx, ny = 30, 21, 11
        t = np.arange(nt) * dt
        x = np.arange(nx) * dx
        y = np.arange(nx) * dy
        f0 = 10
        nfftt = 64
        nfftk = 32
    
        d = np.outer(np.sin(2 * np.pi * f0 * t), np.arange(nx) + 1)
        d = np.tile(d[:, :, np.newaxis], [1, 1, ny])
        
        d = VectorNumpy(d.astype(np.complex))
        F = FFT(model=d, nfft=(nfftt, nfftk, nfftk), sampling=(dt, dx, dy))
        F.dotTest(True)
        
        D = F * d
        dinv = F.H * D
    
        fig, axs = plt.subplots(2, 2, figsize=(10, 6))
        axs[0][0].imshow(d.getNdArray()[:, :, ny // 2].real,
                         vmin=-20, vmax=20, cmap='seismic')
        axs[0][0].set_title('Signal')
        axs[0][0].axis('tight')
        axs[0][1].imshow(np.abs(np.fft.fftshift(D.getNdArray(), axes=1)[:20, :, nfftk // 2]),
                         cmap='seismic')
        axs[0][1].set_title('Fourier Transform')
        axs[0][1].axis('tight')
        axs[1][0].imshow(dinv.getNdArray()[:, :, ny // 2].real,
                         vmin=-20, vmax=20, cmap='seismic')
        axs[1][0].set_title('Inverted')
        axs[1][0].axis('tight')
        axs[1][1].imshow(d.getNdArray()[:, :, ny // 2].real - dinv.getNdArray()[:, :, ny // 2].real,
                         vmin=-20, vmax=20, cmap='seismic')
        axs[1][1].set_title('Error')
        axs[1][1].axis('tight')
        fig.tight_layout()
        plt.show()
    
    print('')



from __future__ import division, print_function, absolute_import
import numpy as np
import cupy as cp
from cupyx.scipy.ndimage import gaussian_filter
try:
    from cusignal.convolution import convolve, correlate
except ModuleNotFoundError:
    raise ModuleNotFoundError("cuSIGNAL is not installed. Please install it")
from occamypy import Operator, Dstack, Vector, superVector
from occamypy.cupy import VectorCupy


class GaussianFilter(Operator):
    def __init__(self, model, sigma):
        """
        Gaussian smoothing operator using scipy smoothing:
        :param model : vector class;
            domain vector
        :param sigma : scalar or sequence of scalars;
            standard deviation along the model directions
        """
        self.sigma = sigma
        self.scaling = np.sqrt(np.prod(np.array(self.sigma) / cp.pi))  # in order to have the max amplitude 1
        
        super(GaussianFilter, self).__init__(model, model)
        return
    
    def __str__(self):
        return "GausFilt"
    
    def forward(self, add, model, data):
        """Forward operator"""
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        # Getting Ndarrays
        model_arr = model.getNdArray()
        data_arr = data.getNdArray()
        data_arr[:] += self.scaling * gaussian_filter(model_arr, sigma=self.sigma)
        return
    
    def adjoint(self, add, model, data):
        """Self-adjoint operator"""
        self.forward(add, data, model)
        return


class ConvND(Operator):
    """
    ND convolution square operator in the domain space

    :param domain   : [no default] - vector class; domain vector
    :param kernel   : [no default] - vector class; kernel vector
    :param method   : [auto] - str; how to compute the convolution [auto, direct, fft]
    :return         : Convolution Operator
    """
    
    def __init__(self, domain, kernel, method='auto'):
        
        if isinstance(kernel, Vector):
            self.kernel = kernel.clone().getNdArray()
        elif isinstance(kernel, cp.ndarray):
            self.kernel = kernel.copy()
        else:
            raise ValueError("kernel has to be either a vector or a cupy.ndarray")
        
        # Padding array to avoid edge effects
        pad_width = []
        for len_filt in self.kernel.shape:
            half_len = int(len_filt / 2)
            if np.mod(len_filt, 2):
                padding = (half_len, half_len)
            else:
                padding = (half_len, half_len - 1)
            pad_width.append(padding)
        self.kernel = cp.padding(self.kernel, pad_width, mode='constant')
        
        if len(domain.shape()) != len(self.kernel.shape):
            raise ValueError("Domain and kernel number of dimensions mismatch")
        
        assert method in ["auto", "direct", "fft"], "method has to be auto, direct or fft"
        self.method = method
        
        super(ConvND, self).__init__(domain, domain)
    
    def __str__(self):
        return " ConvOp "
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        modelNd = model.getNdArray()
        dataNd = data.getNdArray()[:]
        dataNd += convolve(modelNd, self.kernel)
        return
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        modelNd = model.getNdArray()
        dataNd = data.getNdArray()[:]
        modelNd += correlate(dataNd, self.kernel)
        return


def ZeroPad(domain, pad):
    if isinstance(domain, VectorCupy):
        return _ZeroPadIC(domain, pad)
    elif isinstance(domain, superVector):
        # TODO add the possibility to have different padding for each sub-vector
        return Dstack([_ZeroPadIC(v, pad) for v in domain.vecs])
    else:
        raise ValueError("ERROR! Provided domain has to be either vector or superVector")


def _pad_vectorIC(vec, pad):
    if not isinstance(vec, VectorCupy):
        raise ValueError("ERROR! Provided vector must be of vectorCcupy type")
    assert len(vec.shape) == len(pad), "Dimensions of vector and padding mismatch!"
    
    vec_new_shape = tuple(cp.asarray(vec.shape) + [sum(pad[_]) for _ in range(len(pad))])
    if isinstance(vec, VectorCupy):
        return VectorCupy(cp.empty(vec_new_shape, dtype=vec.getNdArray().dtype))
    else:
        raise ValueError("ERROR! For now only vectorCupy is supported!")


class _ZeroPadIC(Operator):
    
    def __init__(self, domain, pad):
        """ Zero Pad operator.

        To pad 2 values to each side of the first dim, and 3 values to each side of the second dim, use:
            pad=((2,2), (3,3))
        :param domain: vectorIC class
        :param pad: scalar or sequence of scalars
            Number of samples to pad in each dimension.
            If a single scalar is provided, it is assigned to every dimension.
        """
        if isinstance(domain, VectorCupy):
            self.dims = domain.shape
            pad = [(pad, pad)] * len(self.dims) if pad is cp.isscalar else list(pad)
            if (cp.array(pad) < 0).any():
                raise ValueError('Padding must be positive or zero')
            self.pad = pad
            super(_ZeroPadIC, self).__init__(domain, _pad_vectorIC(domain, self.pad))
    
    def __str__(self):
        return "ZeroPad "
    
    def forward(self, add, model, data):
        """Zero padding"""
        self.checkDomainRange(model, data)
        if add:
            temp = data.clone()
        y = cp.padding(model.getNdArray(), self.pad, mode='constant')
        data.getNdArray()[:] = y
        if add:
            data.scaleAdd(temp, 1., 1.)
        return
    
    def adjoint(self, add, model, data):
        """Extract non-zero subsequence"""
        self.checkDomainRange(model, data)
        if add:
            temp = model.clone()
        x = data.clone().getNdArray()
        for ax, pad in enumerate(self.pad):
            x = cp.take(x, pad[0] + cp.arange(self.dims[ax]), axis=ax)
        model.arr = x
        if add:
            model.scaleAdd(temp, 1., 1.)
        return



import numpy as np
from scipy.io import loadmat, savemat
from occamypy.vector import Vector, VectorIC, superVector
from occamypy.utils import sep
from occamypy.operator import Operator

# try:
#     from oct2py import Oct2Py
# except ModuleNotFoundError:
#     import subprocess
#     import sys
#     subprocess.call([sys.executable, "-m", "pip", "install", "--user", "oct2py"])
#     from oct2py import Oct2Py
#
# octave = Oct2Py()

# _forward = "function y = Forward(x)\n"\
#            "    y = x-5"\
#            "end"

try:
    import matlab.engine as ME
except ModuleNotFoundError:
    print("MATLAB Engine API not installed."
          "Please find your `matlabroot` and run\n\t"
          "python matlabroot/extern/engines/python/setup.py install")


class Matlab(Operator):
    """Operator built upon a pair of matlab script"""
    
    def __init__(self, domain, range, forward, adjoint):
        """Class constructor
        :param domain  : domain vector
        :param range   : range vector
        :param forward : matlab script for forward operation
        :param adjoint : matlab script for adjoint operation
        """
        if not isinstance(domain, Vector):
            raise TypeError("ERROR! Domain vector not a vector object")
        if not isinstance(range, Vector):
            raise TypeError("ERROR! Range vector not a vector object")
        # Setting domain and range of operator and matrix to use during application of the operator
        self.setDomainRange(domain, range)
        
        eng = ME.start_matlab()
    
    def __str__(self):
        return "MatlabOp"
    
    def forward(self, add, model, data):
        """d = A * m"""
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        model_arr = model.getNdArray()
        data_arr = data.getNdArray()
        matlab_result = eng.
        data_arr +=
        return
    
    def adjoint(self, add, model, data):
        """m = A' * d"""
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        data_arr = data.getNdArray()
        if self.outcore:
            [model_arr, model_axis] = sep.read_file(model.vecfile)
            model_arr += np.matmul(self.M.H, data_arr.ravel()).reshape(model_arr.shape)
            sep.write_file(model.vecfile, model_arr, model_axis)
        else:
            model_arr = model.getNdArray()
            model_arr += np.matmul(self.M.T.conj(), data_arr.ravel()).reshape(model_arr.shape)
        return


if __name__ == '__main__':
    nx = 201
    x = VectorIC((nx,)).zero()
    x.getNdArray()[20:30] = 10.
    x.getNdArray()[50:75] = -5.
    x.getNdArray()[100:150] = 2.5
    x.getNdArray()[175:180] = 7.5
    
    print(0)



import torch
import numpy as np


def set_backends():
    torch.backends.cudnn.enabled = True
    torch.backends.cudnn.benchmark = True
    torch.backends.cudnn.deterministic = True


def set_seed_everywhere(seed=0):
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


__all__ = [
    "set_backends",
    "set_seed_everywhere",
]



from .os import *
from .sep import *
from .logger import *

__all__ = [
    "Logger",
    "read_file",
    "write_file",
    "RunShellCmd",
    "hashfile",
    "mkdir",
    "rand_name",
    "CUPY_ENABLED",
]



import numpy as np
import imageio
from occamypy.vector import Vector


def vector2gif(data: Vector, filename: str, transpose:bool=False, clip:int=100, frame_axis=-1, frame_interval:float=.25):
    """
    Save a 3D vector to an animated GIF file.
    
    :param data: vector - must have 3 dimensions
    :param filename: `str` - path to file to be written with extension
    :param transpose: `bool` - transpose the image axis [False]. Notice that sepvector have the axis swapped w.r.t. numpy
    :param clip: `float` - percentile for clipping the data
    :param frame_axis: `int` - frame axis of data [-1]
    :param frame_interval: `float`- time between frames [0.25]

    """
    assert data.ndim == 3, "Data has to be a 3D vector"
    
    clip_val = np.percentile(np.absolute(data), clip)
    data = np.clip(data, -clip_val, clip_val)
    if frame_axis > 0:  # need to bring the dim. to first dim
        data = np.swapaxes(data, frame_axis, 0)
    if transpose:
        frames = [data[_].T for _ in range(data.shape[0])]
    else:
        frames = [data[_] for _ in range(data.shape[0])]

    imageio.mimsave(filename, frames, 'GIF', **{'duration': frame_interval})




import os
from setuptools import setup, find_packages


def src(pth):
    return os.path.join(os.path.dirname(__file__), pth)


setup(name='occamypy',
      version='0.0.1dev3',
      url="https://github.com/fpicetti/occamypy",  # used for the documentation. TODO switch to readthedocs?
      description='An Object-Oriented Optimization Framework for Large-Scale Inverse Problems',
      long_description=open(src('README.md'), "r").read(),
      long_description_content_type='text/markdown',
      keywords=['algebra', 'inverse problems', 'large-scale optimization'],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Operating System :: Unix'
      ],
      license='GNU',
      author='Ettore Biondi, Guillame Barnier, Robert Clapp, Francesco Picetti, Stuart Farris',
      author_email='francesco.picetti@polimi.it',
      install_requires=['numpy',
                        'scipy',
                        'h5py',
                        'numba',
                        'torch>=1.7.0',
                        'dask',
                        'dask-jobqueue',
                        'dask-kubernetes',
                        'matplotlib',
                        'gputil',
                        ],
      # extras_require={  # one can install two of them with pip install occamypy[cuda]
      #     'cuda': ['cupy>=8.0', 'gputil']},
      packages=find_packages(),
      zip_safe=True)



from math import isnan
from occamypy.vector import superVector
from occamypy import problem as P
from occamypy import operator as O


class NonlinearLeastSquares(P.Problem):
    """Non-linear inverse problem of the form 1/2*|f(m)-d|_2"""
    
    def __init__(self, model, data, op, grad_mask=None,
                 minBound=None, maxBound=None, boundProj=None):
        """
           Constructor of non-linear problem:
           model    	= [no default] - vector class; Initial model vector
           data     	= [no default] - vector class; Data vector
           op       	= [no default] - non-linear operator class; f(m) operator
           grad_mask	= [None] - vector class; Mask to be applied on the gradient during the inversion
           minBound		= [None] - vector class; Minimum value bounds
           maxBound		= [None] - vector class; Maximum value bounds
           boundProj	= [None] - Bounds class; Class with a function "apply(input_vec)" to project input_vec onto some convex set
        """
        # Setting the bounds (if any)
        super(NonlinearLeastSquares, self).__init__(minBound, maxBound, boundProj)
        # Setting internal vector
        self.model = model
        self.dmodel = model.clone()
        self.dmodel.zero()
        # Gradient vector
        self.grad = self.dmodel.clone()
        # Copying the pointer to data vector
        self.data = data
        # Residual vector
        self.res = data.clone()
        self.res.zero()
        # Dresidual vector
        self.dres = self.res.clone()
        # Setting non-linear and linearized operators
        if isinstance(op, O.NonlinearOperator):
            self.op = op
        else:
            raise TypeError("Not provided a non-linear operator!")
        # Checking if a gradient mask was provided
        self.grad_mask = grad_mask
        if self.grad_mask is not None:
            if not grad_mask.checkSame(model):
                raise ValueError("Mask size not consistent with model vector!")
            self.grad_mask = grad_mask.clone()
        # Setting default variables
        self.setDefaults()
        self.linear = False
        return
    
    def __del__(self):
        """Default destructor"""
        return
    
    def resf(self, model):
        """Method to return residual vector r = f(m) - d"""
        self.op.nl_op.forward(False, model, self.res)
        # Computing f(m) - d
        self.res.scaleAdd(self.data, 1., -1.)
        return self.res
    
    def gradf(self, model, res):
        """Method to return gradient vector g = F'r = F'(f(m) - d)"""
        # Setting model point on which the F is evaluated
        self.op.set_background(model)
        # Computing F'r = g
        self.op.lin_op.adjoint(False, self.grad, res)
        # Applying the gradient mask if present
        if self.grad_mask is not None:
            self.grad.multiply(self.grad_mask)
        return self.grad
    
    def dresf(self, model, dmodel):
        """Method to return residual vector dres = Fdm"""
        # Setting model point on which the F is evaluated
        self.op.set_background(model)
        # Computing Fdm = dres
        self.op.lin_op.forward(False, dmodel, self.dres)
        return self.dres
    
    def objf(self, residual):
        """Method to return objective function value 1/2|f(m)-d|_2"""
        val = residual.norm()
        obj = 0.5 * val * val
        return obj


class NonlinearLeastSquaresRegularized(P.Problem):
    """
       Linear inverse problem regularized of the form
            1/2*|f(m)-d|_2 + epsilon^2/2*|Am - m_prior|_2
                or with a non-linear regularization
            1/2*|f(m)-d|_2 + epsilon^2/2*|g(m) - m_prior|_2
    """
    
    def __init__(self, model, data, op, epsilon, grad_mask=None, reg_op=None, prior_model=None,
                 minBound=None, maxBound=None, boundProj=None):
        """
           Constructor of non-linear regularized problem:
           model    	= [no default] - vector class; Initial model vector
           data     	= [no default] - vector class; Data vector
           op       	= [no default] - non-linear operator class; f(m) operator
           epsilon      = [no default] - float; regularization weight
           grad_mask	= [None] - vector class; Mask to be applied on the gradient during the inversion
           reg_op       = [Identity] - non-linear/linear operator class; g(m) regularization operator
           prior_model  = [None] - vector class; Prior model for regularization term
           minBound		= [None] - vector class; Minimum value bounds
           maxBound		= [None] - vector class; Maximum value bounds
           boundProj	= [None] - Bounds class; Class with a function "apply(input_vec)" to project input_vec onto some convex set
        """
        # Setting the bounds (if any)
        super(NonlinearLeastSquaresRegularized, self).__init__(minBound, maxBound, boundProj)
        # Setting internal vector
        self.model = model
        self.dmodel = model.clone()
        self.dmodel.zero()
        # Gradient vector
        self.grad = self.dmodel.clone()
        # Copying the pointer to data vector
        self.data = data
        # Setting a prior model (if any)
        self.prior_model = prior_model
        # Setting linear operators
        # Assuming identity operator if regularization operator was not provided
        if reg_op is None:
            Id_op = O.Identity(self.model)
            reg_op = O.NonlinearOperator(Id_op, Id_op)
        # Checking if space of the prior model is constistent with range of regularization operator
        if self.prior_model is not None:
            if not self.prior_model.checkSame(reg_op.range):
                raise ValueError("Prior model space no constistent with range of regularization operator")
        # Setting non-linear and linearized operators
        if not isinstance(op, O.NonlinearOperator):
            raise TypeError("Not provided a non-linear operator!")
        # Setting non-linear stack of operators
        self.op = O.NonlinearVstack(op, reg_op)
        self.epsilon = epsilon  # Regularization weight
        # Residual vector (data and model residual vectors)
        self.res = self.op.nl_op.range.clone()
        self.res.zero()
        # Dresidual vector
        self.dres = self.res.clone()
        # Checking if a gradient mask was provided
        self.grad_mask = grad_mask
        if self.grad_mask is not None:
            if not grad_mask.checkSame(model):
                raise ValueError("Mask size not consistent with model vector!")
            self.grad_mask = grad_mask.clone()
        # Setting default variables
        self.setDefaults()
        self.linear = False
        # Objective function terms (useful to analyze each term)
        self.obj_terms = [None, None]
        return
    
    def __del__(self):
        """Default destructor"""
        return
    
    def estimate_epsilon(self, verbose=False, logger=None):
        """Method returning epsilon that balances the two terms of the objective function"""
        msg = "Epsilon Scale evaluation"
        if verbose:
            print(msg)
        if logger:
            logger.addToLog("REGULARIZED PROBLEM log file\n" + msg)
        # Keeping the initial model vector
        prblm_mdl = self.get_model()
        # Keeping user-predefined epsilon if any
        epsilon = self.epsilon
        # Setting epsilon to one to evaluate the scale
        self.epsilon = 1.0
        prblm_res = self.get_res(prblm_mdl)  # Compute residual arising from the gradient
        # Balancing the two terms of the objective function
        res_data_norm = prblm_res.vecs[0].norm()
        res_model_norm = prblm_res.vecs[1].norm()
        if isnan(res_model_norm) or isnan(res_data_norm):
            raise ValueError("Obtained NaN: Residual-data-side-norm = %s, Residual-model-side-norm = %s"
                             % (res_data_norm, res_model_norm))
        if res_model_norm == 0.:
            msg = "Trying to perform a linearized step"
            if verbose:
                print(msg)
            prblm_grad = self.get_grad(prblm_mdl)  # Compute first gradient
            # Gradient in the data space
            prblm_dgrad = self.get_dres(prblm_mdl, prblm_grad)
            # Computing linear step length
            dgrad0_res = prblm_res.vecs[0].dot(prblm_dgrad.vecs[0])
            dgrad0_dgrad0 = prblm_dgrad.vecs[0].dot(prblm_dgrad.vecs[0])
            if isnan(dgrad0_res) or isnan(dgrad0_dgrad0):
                raise ValueError("Obtained NaN: gradient-dataspace-norm = %s, gradient-dataspace-dot-residuals = %s"
                                 % (dgrad0_dgrad0, dgrad0_res))
            if dgrad0_dgrad0 != 0.:
                alpha = -dgrad0_res / dgrad0_dgrad0
            else:
                msg = "Cannot compute linearized alpha for the given problem! Provide a different initial model"
                if logger:
                    logger.addToLog(msg)
                raise ValueError(msg)
            # model=model+alpha*grad
            prblm_mdl.scaleAdd(prblm_grad, 1.0, alpha)
            prblm_res = self.resf(prblm_mdl)
            # Recompute the new objective function terms
            res_data_norm = prblm_res.vecs[0].norm()
            res_model_norm = prblm_res.vecs[1].norm()
            # If regularization term is still zero, stop the solver
            if res_model_norm == 0.:
                msg = "Model residual component norm is zero, cannot find epsilon scale! Provide a different initial model"
                if logger:
                    logger.addToLog(msg)
                raise ValueError(msg)
        # Resetting user-predefined epsilon if any
        self.epsilon = epsilon
        epsilon_balance = res_data_norm / res_model_norm
        # Setting default variables
        self.setDefaults()
        self.linear = False
        msg = "	Epsilon balancing the the two objective function terms is: %.2e" % epsilon_balance
        if verbose:
            print(msg)
        if logger:
            logger.addToLog(msg + "\nREGULARIZED PROBLEM end log file")
        return epsilon_balance
    
    def resf(self, model):
        """
        Method to return residual vector r = [r_d; r_m]:
        r_d = f(m) - d;
        r_m = Am - m_prior or r_m = g(m) - m_prior
        """
        self.op.nl_op.forward(False, model, self.res)
        # Computing r_d = f(m) - d
        self.res.vecs[0].scaleAdd(self.data, 1., -1.)
        # Computing r_m = Am - m_prior
        if self.prior_model is not None:
            self.res.vecs[1].scaleAdd(self.prior_model, 1., -1.)
        # Scaling by epsilon epsilon*r_m
        self.res.vecs[1].scale(self.epsilon)
        return self.res
    
    def gradf(self, model, res):
        """
        Method to return gradient vector
        g = F'r_d + (epsilon*A'r_m or epsilon*G'r_m)
        """
        # Setting model point on which the F is evaluated
        self.op.set_background(model)
        # g = epsilon*A'r_m
        self.op.lin_op.ops[1].adjoint(False, self.grad, res.vecs[1])
        self.grad.scale(self.epsilon)
        # g = F'r_d + A'(epsilon*r_m)
        self.op.lin_op.ops[0].adjoint(True, self.grad, res.vecs[0])
        # Applying the gradient mask if present
        if self.grad_mask is not None:
            self.grad.multiply(self.grad_mask)
        return self.grad
    
    def dresf(self, model, dmodel):
        """
        Method to return residual vector
        dres = [F + epsilon * (A or G)]dm
        """
        # Setting model point on which the F is evaluated
        self.op.set_background(model)
        # Computing Ldm = dres_d
        self.op.lin_op.forward(False, dmodel, self.dres)
        # Scaling by epsilon
        self.dres.vecs[1].scale(self.epsilon)
        return self.dres
    
    def objf(self, residual):
        """
        Method to return objective function value
        1/2|f(m)-d|_2 + (epsilon^2/2*|Am-m_prior|_2 or epsilon^2/2*|g(m)-m_prior|_2)
        """
        # data term
        val = residual.vecs[0].norm()
        self.obj_terms[0] = 0.5 * val * val
        # model term
        val = residual.vecs[1].norm()
        self.obj_terms[1] = 0.5 * val * val
        obj = self.obj_terms[0] + self.obj_terms[1]
        return obj


class VarProRegularized(P.Problem):
    """
    Non-linear inverse problem in which part of the model parameters define a quadratic function
    The non-linear component is solved using the variable-projection method (Golub and Pereyra, 1973)
    Problem form: phi(m) = 1/2*|g(m_nl) + h(m_nl)m_lin - d|_2 + epsilon^2/2*|g'(m_nl) + h'(m_nl)m_lin - d'|_2
    """
    
    def __init__(self, model_nl, lin_model, h_op, data, lin_solver, g_op=None, g_op_reg=None, h_op_reg=None,
                 data_reg=None, epsilon=None, minBound=None, maxBound=None, boundProj=None, prec=None, warm_start=False):
        """
            Constructor for solving a inverse problem using the variable-projection method
            Required arguments:
            model_nl    = [no default] - vector class; Initial non-linear model component of the objective function
            lin_model   = [no default] - vector class; Initial quadritic (Linear) model component of the objective function (will be zeroed out)
            h_op   		= [no default] - Vp operator class; Variable projection operator
            data   		= [no default] - vector class; Data vector
            lin_solver	= [no default] - solver class; Linear solver to invert for linear component of the model
            Optional arguments:
            g_op   		= [None] - non-linear operator class; Fully non-linear additional operator
            g_op_reg   	= [None] - non-linear operator class; Fully non-linear additional operator for regularization term
            h_op_reg	= [None] - Vp operator class; Variable projection operator for regularization term
            data_reg   	= [None] - vector class; Data vector for regularization term
            epsilon 	= [None] - float; Regularization term weight (must be provided if a regularization is needed)
            minBound	= [None] - vector class; Minimum value bounds
            maxBound	= [None] - vector class; Maximum value bounds
            boundProj	= [None] - Bounds class; Class with a function "apply(input_vec)" to project input_vec onto some convex set
            prec       	= [None] - linear operator class; Preconditioning matrix for VP problem
            warm_start  = [None] - boolean; Start VP problem from previous linearly inverted model
            ####################################################################################################################################
            Note that to save the results of the linear inversion the user has to specify the saving parameters within the setDefaults of the
            linear solver. The results can only be saved on files. To the prefix specified within the lin_solver f_eval_# will be added.
        """
        if not isinstance(h_op, O.VarProOperator):
            raise TypeError("ERROR! Not provided an operator class for the variable projection problem")
        # Setting the bounds (if any)
        super(VarProRegularized, self).__init__(minBound, maxBound, boundProj)
        # Setting internal vector
        self.model = model_nl
        self.dmodel = model_nl.clone()
        self.dmodel.zero()
        # Linear component of the inverted model
        self.lin_model = lin_model
        self.lin_model.zero()
        # Copying the pointer to data vector
        self.data = data
        # Setting non-linear/linear operator
        if not isinstance(h_op, O.VarProOperator):
            raise TypeError("ERROR! Provide a VpOperator operator class for h_op")
        self.h_op = h_op
        # Setting non-linear operator (if any)
        self.g_op = g_op
        # Verifying if a regularization is requested
        self.epsilon = epsilon
        # Setting non-linear regularization operator
        self.g_op_reg = g_op_reg
        # Setting non-linear/linear operator
        self.h_op_reg = h_op_reg
        # Setting data term in regularization
        self.data_reg = data_reg
        if self.h_op_reg is not None and self.epsilon is None:
            raise ValueError("ERROR! Epsilon value must be provided if a regularization term is requested.")
        # Residual vector
        if self.epsilon is not None:
            # Creating regularization residual vector
            res_reg = None
            if self.g_op_reg is not None:
                res_reg = self.g_op_reg.nl_op.range.clone()
            elif self.h_op_reg is not None:
                if not isinstance(h_op_reg, O.VarProOperator):
                    raise TypeError("ERROR! Provide a VpOperator operator class for h_op_reg")
                res_reg = self.h_op_reg.h_lin.range.clone()
            elif self.data_reg is not None:
                res_reg = self.data_reg.clone()
            # Checking if a residual vector for the regularization term was created
            if res_reg is None:
                raise ValueError("ERROR! If epsilon is provided, then a regularization term must be provided")
            self.res = superVector(data.clone(), res_reg)
            # Objective function terms (useful to analyze each term)
            self.obj_terms = [None, None]
        else:
            self.res = data.clone()
        # Instantiating linear inversion problem
        if self.h_op_reg is not None:
            self.vp_linear_prob = P.LeastSquaresRegularized(self.lin_model, self.data, self.h_op.h_lin, self.epsilon,
                                                            reg_op=self.h_op_reg.h_lin, prior_model=self.data_reg,
                                                            prec=prec)
        else:
            self.vp_linear_prob = P.LeastSquares(self.lin_model, self.data, self.h_op.h_lin, prec=prec)
        # Zeroing out the residual vector
        self.res.zero()
        # Dresidual vector
        self.dres = self.res.clone()
        # Gradient vector
        self.grad = self.dmodel.clone()
        # Setting default variables
        self.setDefaults()
        self.linear = False
        # Linear solver for inverting quadratic component
        self.lin_solver = lin_solver
        self.lin_solver.flush_memory = True
        self.lin_solver_prefix = self.lin_solver.prefix
        self.vp_linear_prob.linear = True
        self.warm_start = warm_start
        return
    
    def __del__(self):
        """Default destructor"""
        return
    
    def estimate_epsilon(self, verbose=False, logger=None):
        """Method returning epsilon that balances the two terms of the objective function"""
        if self.epsilon is None:
            raise ValueError("ERROR! Problem is not regularized, cannot evaluate epsilon value!")
        if self.g_op_reg is not None and self.h_op_reg is None:
            # Problem is non-linearly regularized
            msg = "Epsilon Scale evaluation"
            if verbose: print(msg)
            if logger: logger.addToLog("REGULARIZED PROBLEM log file\n" + msg)
            # Keeping the initial model vector
            prblm_mdl = self.get_model()
            # Keeping user-predefined epsilon if any
            epsilon = self.epsilon
            # Setting epsilon to one to evaluate the scale
            self.epsilon = 1.0
            prblm_res = self.get_res(prblm_mdl)  # Compute residual arising from the gradient
            # Balancing the two terms of the objective function
            res_data_norm = prblm_res.vecs[0].norm()
            res_model_norm = prblm_res.vecs[1].norm()
            if isnan(res_model_norm) or isnan(res_data_norm):
                raise ValueError("ERROR! Obtained NaN: Residual-data-side-norm = %s, Residual-model-side-norm = %s" % (
                    res_data_norm, res_model_norm))
            if res_model_norm == 0.0:
                msg = "Model residual component norm is zero, cannot find epsilon scale! Provide a different initial model"
                if (logger): logger.addToLog(msg)
                raise ValueError(msg)
            # Resetting user-predefined epsilon if any
            self.epsilon = epsilon
            epsilon_balance = res_data_norm / res_model_norm
            # Resetting problem
            self.setDefaults()
            msg = "	Epsilon balancing the the two objective function terms is: %s" % (epsilon_balance)
            if verbose: print(msg)
            if logger: logger.addToLog(msg + "\nREGULARIZED PROBLEM end log file")
        elif self.h_op_reg is not None:
            # Setting non-linear component of the model
            self.h_op.set_nl(self.model)
            self.h_op_reg.set_nl(self.model)
            # Problem is linearly regularized (fixing non-linear part and evaluating the epsilon on the linear
            # component)
        return self.vp_linear_prob.estimate_epsilon(verbose, logger)
    
    def resf(self, model):
        """Method to return residual vector"""
        # Zero-out residual vector
        self.res.zero()
        ###########################################
        # Applying full non-linear modeling operator
        res = self.res
        if self.epsilon is not None: res = self.res.vecs[0]
        # Computing non-linear part g(m) (if any)
        if self.g_op is not None:
            self.g_op.nl_op.forward(False, model, res)
        # Computing non-linear part g_reg(m) (if any)
        if self.g_op_reg is not None:
            self.g_op_reg.nl_op.forward(False, model, self.res.vecs[1])
        
        ##################################
        # Setting data for linear inversion
        # data term = data - [g(m) if any]
        res.scaleAdd(self.data, -1.0, 1.0)
        # Setting data within first term
        self.vp_linear_prob.data = res
        
        # regularization data term = [g_reg(m) - data_reg if any]
        if self.data_reg is not None:
            self.res.vecs[1].scaleAdd(self.data_reg, 1.0, -1.0)
        # Data term for linear regularization term
        if "epsilon" in dir(self.vp_linear_prob):
            self.res.vecs[1].scale(-1.0)
            self.vp_linear_prob.prior_model = self.res.vecs[1]
        
        ##################################
        # Running linear inversion
        # Getting fevals for saving linear inversion results
        fevals = self.get_fevals()
        # Setting initial linear inversion model
        if not self.warm_start:
            self.lin_model.zero()
        self.vp_linear_prob.set_model(self.lin_model)
        # Setting non-linear component of the model
        self.h_op.set_nl(model)
        if self.h_op_reg is not None:
            self.h_op_reg.set_nl(model)
        # Resetting inversion problem variables
        self.vp_linear_prob.setDefaults()
        # Saving linear inversion results if requested
        if self.lin_solver_prefix is not None:
            self.lin_solver.setPrefix(self.lin_solver_prefix + "_feval%s" % fevals)
        
        # Printing non-linear inversion information
        if self.lin_solver.logger is not None:
            # Writing linear inversion log information if requested (i.e., a logger is present in the solver)
            msg = "NON_LINEAR INVERSION INFO:\n	objective function evaluation\n"
            msg += "#########################################################################################\n"
            self.lin_solver.logger.addToLog(msg + "Linear inversion for non-linear function evaluation # %s" % (fevals))
        self.lin_solver.run(self.vp_linear_prob, verbose=False)
        if self.lin_solver.logger is not None:
            self.lin_solver.logger.addToLog(
                "#########################################################################################\n")
        # Copying inverted linear optimal model
        self.lin_model.copy(self.vp_linear_prob.get_model())
        # Flushing internal saved results of the linear inversion
        self.lin_solver.flush_results()
        
        ##################################
        # Obtaining the residuals
        if (self.epsilon is not None) and not ("epsilon" in dir(self.vp_linear_prob)):
            # Regularization contains a non-linear operator only
            self.res.vecs[0].copy(self.vp_linear_prob.get_res(self.lin_model))
            self.res.vecs[1].scale(self.epsilon)
        else:
            self.res.copy(self.vp_linear_prob.get_res(self.lin_model))
        return self.res
    
    def gradf(self, model, res):
        """
           Method to return gradient vector
           grad= [G(m)' + H(m_nl;m_lin)'] r_d + epsilon * [G'(m_nl)' + H'(m_nl;m_lin)'] r_m
        """
        # Zero-out gradient vector
        self.grad.zero()
        # Setting the optimal linear model component and background of the Jacobian matrices
        self.h_op.set_lin_jac(self.lin_model)  # H(_,m_lin_opt)
        self.h_op.h_nl.set_background(model)  # H(m_nl,m_lin_opt)
        if self.h_op_reg is not None:
            self.h_op_reg.set_lin_jac(self.lin_model)  # H'(_,m_lin_opt)
            self.h_op_reg.h_nl.set_background(model)  # H'(m_nl,m_lin_opt)
        if self.g_op is not None:
            self.g_op.set_background(model)  # G(m_nl)
        if self.g_op_reg is not None:
            self.g_op_reg.set_background(model)  # G'(m_nl)
        # Computing contribuition from the regularization term (if any)
        if self.epsilon is not None:
            # G'(m_nl)' r_m
            if self.g_op_reg is not None:
                self.g_op_reg.lin_op.adjoint(False, self.grad, res.vecs[1])
            # H'(m_nl,m_lin_opt)' r_m
            if self.h_op_reg is not None:
                self.h_op_reg.h_nl.lin_op.adjoint(True, self.grad, res.vecs[1])
            # epsilon * [G'(m_nl)' + H'(m_nl,m_lin_opt)'] r_m
            self.grad.scale(self.epsilon)
        res = self.res if self.epsilon is None else self.res.vecs[0]
        # G(m_nl)' r_d
        if self.g_op is not None:
            self.g_op.lin_op.adjoint(True, self.grad, res)
        # H(m_nl,m_lin_opt)' r_d
        self.h_op.h_nl.lin_op.adjoint(True, self.grad, res)
        if self.lin_solver.logger is not None:
            self.lin_solver.logger.addToLog(
                "NON_LINEAR INVERSION INFO:\n	Gradient has been evaluated, current objective function value: %s;\n 	"
                "Stepping!" % (
                    self.get_obj(model)))
        return self.grad
    
    def dresf(self, model, dmodel):
        """Method to return residual vector dres (Not currently supported)"""
        raise NotImplementedError(
            "ERROR! dresf is not currently supported! Provide an initial step-length value different than zero.")
    
    def objf(self, residual):
        """
        Method to return objective function value
        1/2*|g(m_nl) + h(m_nl)m_lin - d|_2 + epsilon^2/2*|g'(m_nl) + h'(m_nl)m_lin - d'|_2'
        """
        if "obj_terms" in dir(self):
            # data term
            val = residual.vecs[0].norm()
            self.obj_terms[0] = 0.5 * val * val
            # model term
            val = residual.vecs[1].norm()
            self.obj_terms[1] = 0.5 * val * val
            obj = self.obj_terms[0] + self.obj_terms[1]
        else:
            val = residual.norm()
            obj = 0.5 * val * val
        return obj



from math import isnan

from .base import Problem
from occamypy import operator as O
from occamypy.vector import superVector


class LeastSquares(Problem):
    """Linear inverse problem of the form 1/2*|Lm-d|_2"""

    def __init__(self, model, data, op, grad_mask=None, prec=None,
                 minBound=None, maxBound=None, boundProj=None):
        """
           Constructor of linear problem:
           model    	= [no default] - vector class; Initial model vector
           data     	= [no default] - vector class; Data vector
           op       	= [no default] - linear operator class; L operator
           grad_mask	= [None] - vector class; Mask to be applied on the gradient during the inversion
           minBound     = [None] - vector class; Minimum value bounds
           maxBound     = [None] - vector class; Maximum value bounds
           boundProj	= [None] - Bounds class; Class with a function "apply(input_vec)" to project input_vec onto some convex set
           prec       	= [None] - linear operator class; Preconditioning matrix
        """
        # Setting the bounds (if any)
        super(LeastSquares, self).__init__(minBound, maxBound, boundProj)
        # Setting internal vector
        self.model = model
        self.dmodel = model.clone()
        self.dmodel.zero()
        # Gradient vector
        self.grad = self.dmodel.clone()
        # Copying the pointer to data vector
        self.data = data
        # Residual vector
        self.res = data.clone()
        self.res.zero()
        # Dresidual vector
        self.dres = self.res.clone()
        # Setting linear operator
        self.op = op
        # Checking if a gradient mask was provided
        self.grad_mask = grad_mask
        if self.grad_mask is not None:
            if not grad_mask.checkSame(model):
                raise ValueError("Mask size not consistent with model vector!")
            self.grad_mask = grad_mask.clone()
        # Preconditioning matrix
        self.prec = prec
        # Setting default variables
        self.setDefaults()
        self.linear = True
        return

    def __del__(self):
        """Default destructor"""
        return

    def resf(self, model):
        """Method to return residual vector r = Lm - d"""
        # Computing Lm
        if model.norm() != 0.:
            self.op.forward(False, model, self.res)
        else:
            self.res.zero()
        # Computing Lm - d
        self.res.scaleAdd(self.data, 1., -1.)
        return self.res

    def gradf(self, model, res):
        """Method to return gradient vector g = L'r = L'(Lm - d)"""
        # Computing L'r = g
        self.op.adjoint(False, self.grad, res)
        # Applying the gradient mask if present
        if self.grad_mask is not None:
            self.grad.multiply(self.grad_mask)
        return self.grad

    def dresf(self, model, dmodel):
        """Method to return residual vector dres = Ldm"""
        # Computing Ldm = dres
        self.op.forward(False, dmodel, self.dres)
        return self.dres

    def objf(self, residual):
        """Method to return objective function value 1/2|Lm-d|_2"""
        val = residual.norm()
        obj = 0.5 * val * val
        return obj


class LeastSquaresSymmetric(Problem):
    """Linear inverse problem of the form 1/2m'Am - m'b"""

    def __init__(self, model, data, op, prec=None,
                 minBound=None, maxBound=None, boundProj=None):
        """
        Constructor of linear symmetric problem:
        model    	= [no default] - vector class; Initial model vector
        data     	= [no default] - vector class; Data vector
        op       	= [no default] - linear operator class; A symmetric operator (i.e., A = A')
        minBound		= [None] - vector class; Minimum value bounds
        maxBound		= [None] - vector class; Maximum value bounds
        boundProj	= [None] - Bounds class; Class with a function "apply(input_vec)" to project input_vec onto some convex set
        prec       	= [None] - linear operator class; Preconditioning matrix
        """
        # Setting the bounds (if any)
        super(LeastSquaresSymmetric, self).__init__(minBound, maxBound, boundProj)
        # Checking range and domain are the same
        if not model.checkSame(data) and not op.domain.checkSame(op.range):
            raise ValueError("Data and model vector live in different spaces!")
        # Setting internal vector
        self.model = model
        self.dmodel = model.clone()
        self.dmodel.zero()
        # Copying the pointer to data vector
        self.data = data
        # Residual vector
        self.res = data.clone()
        self.res.zero()
        # Gradient vector is equal to the residual vector
        self.grad = self.res
        # Dresidual vector
        self.dres = self.res.clone()
        # Setting linear operator
        self.op = op
        # Preconditioning matrix
        self.prec = prec
        # Setting default variables
        self.setDefaults()
        self.linear = True

    def __del__(self):
        """Default destructor"""
        return

    def resf(self, model):
        """Method to return residual vector r = Am - b"""
        # Computing Lm
        if model.norm() != 0.:
            self.op.forward(False, model, self.res)
        else:
            self.res.zero()
        # Computing Lm - d
        self.res.scaleAdd(self.data, 1., -1.)
        return self.res

    def gradf(self, model, res):
        """Method to return gradient vector equal to residual one"""
        # Assigning g = r
        self.grad = self.res
        return self.grad

    def dresf(self, model, dmodel):
        """Method to return residual vector dres = Adm"""
        # Computing Ldm = dres
        self.op.forward(False, dmodel, self.dres)
        return self.dres

    def objf(self, residual):
        """Method to return objective function value 1/2m'Am - m'b"""
        obj = 0.5 * (self.model.dot(residual) - self.model.dot(self.data))
        return obj


class LeastSquaresRegularized(Problem):
    """Linear inverse problem regularized of the form 1/2*|Lm-d|_2 + epsilon^2/2*|Am-m_prior|_2"""

    def __init__(self, model, data, op, epsilon, grad_mask=None, reg_op=None, prior_model=None, prec=None,
                 minBound=None, maxBound=None, boundProj=None):
        """
        Constructor of linear regularized problem:
        model    	= [no default] - vector class; Initial model vector
        data     	= [no default] - vector class; Data vector
        op       	= [no default] - linear operator class; L operator
        epsilon      = [no default] - float; regularization weight
        grad_mask	= [None] - vector class; Mask to be applied on the gradient during the inversion
        reg_op       = [Identity] - linear operator class; A regularization operator
        prior_model  = [None] - vector class; Prior model for regularization term
        minBound		= [None] - vector class; Minimum value bounds
        maxBound		= [None] - vector class; Maximum value bounds
        boundProj	= [None] - Bounds class; Class with a function "apply(input_vec)" to project input_vec onto some convex set
        prec       	= [None] - linear operator class; Preconditioning matrix
        """
        # Setting the bounds (if any)
        super(LeastSquaresRegularized, self).__init__(minBound, maxBound, boundProj)
        # Setting internal vector
        self.model = model
        self.dmodel = model.clone()
        self.dmodel.zero()
        # Gradient vector
        self.grad = self.dmodel.clone()
        # Copying the pointer to data vector
        self.data = data
        # Setting a prior model (if any)
        self.prior_model = prior_model
        # Setting linear operators
        # Assuming identity operator if regularization operator was not provided
        if reg_op is None:
            reg_op = O.Identity(self.model)
        # Checking if space of the prior model is consistent with range of
        # regularization operator
        if self.prior_model is not None:
            if not self.prior_model.checkSame(reg_op.range):
                raise ValueError("Prior model space no consistent with range of regularization operator")
        self.op = O.Vstack(op, reg_op)  # Modeling operator
        self.epsilon = epsilon  # Regularization weight
        # Checking if a gradient mask was provided
        self.grad_mask = grad_mask
        if self.grad_mask is not None:
            if not grad_mask.checkSame(model):
                raise ValueError("Mask size not consistent with model vector!")
            self.grad_mask = grad_mask.clone()
        # Residual vector (data and model residual vectors)
        self.res = self.op.range.clone()
        self.res.zero()
        # Dresidual vector
        self.dres = self.res.clone()
        # Setting default variables
        self.setDefaults()
        self.linear = True
        # Preconditioning matrix
        self.prec = prec
        # Objective function terms (useful to analyze each term)
        self.obj_terms = [None, None]

    def __del__(self):
        """Default destructor"""
        return
    
    # TODO make estimate_epsilon a function that takes a problem as input
    def estimate_epsilon(self, verbose=False, logger=None):
        """
        Method returning epsilon that balances the first gradient in the 'extended-data' space or initial data residuals
        """
        msg = "Epsilon Scale evaluation"
        if verbose:
            print(msg)
        if logger:
            logger.addToLog("REGULARIZED PROBLEM log file\n" + msg)
        # Keeping the initial model vector
        prblm_mdl = self.get_model()
        mdl_tmp = prblm_mdl.clone()
        # Keeping user-predefined epsilon if any
        epsilon = self.epsilon
        # Setting epsilon to one to evaluate the scale
        self.epsilon = 1.0
        if self.model.norm() != 0.:
            prblm_res = self.get_res(self.model)
            msg = "	Epsilon balancing data and regularization residuals is: %.2e"
        else:
            prblm_grad = self.get_grad(self.model)  # Compute first gradient
            prblm_res = self.get_res(prblm_grad)  # Compute residual arising from the gradient
            # Balancing the first gradient in the 'extended-data' space
            prblm_res.vecs[0].scaleAdd(self.data)  # Remove data vector (Lg0 - d + d)
            if self.prior_model is not None:
                prblm_res.vecs[1].scaleAdd(self.prior_model)  # Remove prior model vector (Ag0 - m_prior + m_prior)
            msg = "	Epsilon balancing the data-space gradients is: %.2e"
        res_data_norm = prblm_res.vecs[0].norm()
        res_model_norm = prblm_res.vecs[1].norm()
        if isnan(res_model_norm) or isnan(res_data_norm):
            raise ValueError("Obtained NaN: Residual-data-side-norm = %.2e, Residual-model-side-norm = %.2e"
                             % (res_data_norm, res_model_norm))
        if res_model_norm == 0.:
            raise ValueError("Model residual component norm is zero, cannot find epsilon scale")
        # Resetting user-predefined epsilon if any
        self.epsilon = epsilon
        # Resetting problem initial model vector
        self.set_model(mdl_tmp)
        del mdl_tmp
        epsilon_balance = res_data_norm / res_model_norm
        # Resetting feval
        self.fevals = 0
        msg = msg % epsilon_balance
        if verbose:
            print(msg)
        if logger:
            logger.addToLog(msg + "\nREGULARIZED PROBLEM end log file")
        return epsilon_balance

    def resf(self, model):
        """Method to return residual vector r = [r_d; r_m]: r_d = Lm - d; r_m = epsilon * (Am - m_prior) """
        if model.norm() != 0.:
            self.op.forward(False, model, self.res)
        else:
            self.res.zero()
        # Computing r_d = Lm - d
        self.res.vecs[0].scaleAdd(self.data, 1., -1.)
        # Computing r_m = Am - m_prior
        if self.prior_model is not None:
            self.res.vecs[1].scaleAdd(self.prior_model, 1., -1.)
        # Scaling by epsilon epsilon*r_m
        self.res.vecs[1].scale(self.epsilon)
        return self.res

    def gradf(self, model, res):
        """Method to return gradient vector g = L'r_d + epsilon*A'r_m"""
        # Scaling by epsilon the model residual vector (saving temporarily residual regularization)
        # g = epsilon*A'r_m
        self.op.ops[1].adjoint(False, self.grad, res.vecs[1])
        self.grad.scale(self.epsilon)
        # g = L'r_d + epsilon*A'r_m
        self.op.ops[0].adjoint(True, self.grad, res.vecs[0])
        # Applying the gradient mask if present
        if self.grad_mask is not None:
            self.grad.multiply(self.grad_mask)
        return self.grad

    def dresf(self, model, dmodel):
        """Method to return residual vector dres = (L + epsilon * A)dm"""
        # Computing Ldm = dres_d
        self.op.forward(False, dmodel, self.dres)
        # Scaling by epsilon
        self.dres.vecs[1].scale(self.epsilon)
        return self.dres

    def objf(self, residual):
        """Method to return objective function value 1/2|Lm-d|_2 + epsilon^2/2*|Am-m_prior|_2"""
        for idx in range(residual.n):
            val = residual.vecs[idx].norm()
            self.obj_terms[idx] = 0.5 * val*val
        return sum(self.obj_terms)


class Lasso(Problem):
    """Convex problem 1/2*| y - Am |_2 + lambda*| m |_1"""

    def __init__(self, model, data, op, op_norm=None, lambda_value=None,
                 minBound=None, maxBound=None, boundProj=None):
        """
           Constructor of convex L1-norm LASSO inversion problem:
           model    	= [no default] - vector class; Initial model vector
           data     	= [no default] - vector class; Data vector
           op       	= [no default] - linear operator class; L operator
           lambda_value	= [None] - Regularization weight. Not necessary for ISTC solver but required for ISTA and FISTA
           op_norm		= [None] - float; A operator norm that will be evaluated with the power method if not provided
           minBound		= [None] - vector class; Minimum value bounds
           maxBound		= [None] - vector class; Maximum value bounds
           boundProj	= [None] - Bounds class; Class with a function "apply(input_vec)" to project input_vec onto some convex set
        """
        # Setting the bounds (if any)
        super(Lasso, self).__init__(minBound, maxBound, boundProj)
        # Setting internal vector
        self.model = model
        self.dmodel = model.clone()
        self.dmodel.zero()
        # Gradient vector
        self.grad = self.dmodel.clone()
        # Copying the pointer to data vector
        self.data = data
        # Setting linear operator
        self.op = op  # Modeling operator
        # Residual vector (data and model residual vectors)
        self.res = superVector(op.range.clone(), op.domain.clone())
        self.res.zero()
        # Dresidual vector
        self.dres = None  # Not necessary for the inversion
        # Setting default variables
        self.setDefaults()
        self.linear = True
        if op_norm is not None:
            # Using user-provided A operator norm
            self.op_norm = op_norm  # Operator Norm necessary for solver
        else:
            # Evaluating operator norm using power method
            self.op_norm = self.op.powerMethod()
        self.lambda_value = lambda_value
        # Objective function terms (useful to analyze each term)
        self.obj_terms = [None, None]
        return

    def set_lambda(self, lambda_in):
        # Set lambda
        self.lambda_value = lambda_in
        return

    def objf(self, residual):
        """Method to return objective function value 1/2*| y - Am |_2 + lambda*| m |_1"""
        # data term
        val = residual.vecs[0].norm()
        self.obj_terms[0] = 0.5 * val*val
        # model term
        self.obj_terms[1] = self.lambda_value * residual.vecs[1].norm(1)
        return sum(self.obj_terms)

    # define function that computes residuals
    def resf(self, model):
        """ y - alpha * A m = rd (self.res[0]) and m = rm (self.res[1]);"""
        if model.norm() != 0.:
            self.op.forward(False, model, self.res.vecs[0])
        else:
            self.res.zero()
        # Computing r_d = Lm - d
        self.res.vecs[0].scaleAdd(self.data, -1., 1.)
        # Run regularization part
        self.res.vecs[1].copy(model)
        return self.res

    # function that projects search direction into data space (Not necessary for ISTC)
    def dresf(self, model, dmodel):
        """Linear projection of the model perturbation onto the data space. Method not implemented"""
        raise NotImplementedError("dresf is not necessary for ISTC; DO NOT CALL THIS METHOD")

    # function to compute gradient (Soft thresholding applied outside in the solver)
    def gradf(self, model, res):
        """- A'r_data (residual[0]) = g"""
        # Apply an adjoint modeling
        self.op.adjoint(False, self.grad, res.vecs[0])
        # Applying negative scaling
        self.grad.scale(-1.0)
        return self.grad


class GeneralizedLasso(Problem):
    def __init__(self, model, data, op, eps=1., reg=None,
                 minBound=None, maxBound=None, boundProj=None):
        """
        Linear Problem with L1 regularizer:

        .. math ::
            1 / 2 |Op m - d|_2^2 + eps |R m|_1

        :param model        : vector; initial model
        :param data         : vector; data
        :param op           : LinearOperator; modeling operator
        :param eps          : float; weight of L1 regularizer [1.]
        :param reg          : LinearOperator; L1 regularizer [Identity]
        :param minBound     : vector; minimum value bounds
        :param maxBound     : vector; maximum value bounds
        :param boundProj    : Bounds; object with a method "apply(x)" to project x onto some convex set
        """
        super(GeneralizedLasso, self).__init__(minBound, maxBound, boundProj)
        self.model = model
        self.dmodel = model.clone().zero()
        self.grad = self.dmodel.clone()
        self.data = data
        self.op = op

        self.minBound = minBound
        self.maxBound = maxBound
        self.boundProj = boundProj

        # L1 Regularization
        self.reg_op = reg if reg is not None else O.Identity(model)
        self.eps = eps

        # Last settings
        self.obj_terms = [None] * 2
        self.linear = True
        # store the "residuals" (for computing the objective function)
        self.res_data = self.op.range.clone().zero()
        self.res_reg = self.reg_op.range.clone().zero()
        # this last superVector is instantiated with pointers to res_data and res_reg!
        self.res = superVector(self.res_data, self.res_reg)

    def __del__(self):
        """Default destructor"""
        return

    def objf(self, residual, eps=None):
        """
        Compute objective function based on the residual (super)vector

        .. math ::
            1 / 2 |Op m - d|_2^2 + eps |R m|_1

        """
        res_data = residual.vecs[0]
        res_reg = residual.vecs[1]
        eps = eps if eps is not None else self.eps

        # data fidelity
        self.obj_terms[0] = .5 * res_data.norm(2) ** 2

        # regularization penalty
        self.obj_terms[1] = eps * res_reg.norm(1)

        return sum(self.obj_terms)

    def resf(self, model):
        """Compute residuals from current model"""

        # compute data residual: Op * m - d
        if model.norm() != 0:
            self.op.forward(False, model, self.res_data)  # rd = Op * m
        else:
            self.res_data.zero()
        self.res_data.scaleAdd(self.data, 1., -1.)  # rd = rd - d

        # compute L1 reg residuals
        if model.norm() != 0. and self.reg_op is not None:
            self.reg_op.forward(False, model, self.res_reg)
        else:
            self.res_reg.zero()

        return self.res



import numpy as np
from occamypy import Vector, Operator


class Matrix(Operator):
    """Operator built upon a matrix"""
    
    def __init__(self, matrix, domain, range, outcore=False):
        """Class constructor
        :param matrix   : matrix to use
        :param domain   : domain vector
        :param range    : range vector
        :param outcore  : use outcore sep operators
        """
        if not isinstance(domain, Vector):
            raise TypeError("ERROR! Domain vector not a vector object")
        if not isinstance(range, Vector):
            raise TypeError("ERROR! Range vector not a vector object")
        # Setting domain and range of operator and matrix to use during application of the operator
        super().__init__(domain, range)
        if not isinstance(matrix, np.ndarray):
            raise ValueError("ERROR! matrix has to be a numpy ndarray")
        self.M = matrix
        self.outcore = outcore
    
    def __str__(self):
        return "MatrixOp"
    
    def forward(self, add, model, data):
        """d = A * m"""
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        model_arr = model.getNdArray()
        data_arr = data.getNdArray()
        data_arr += np.matmul(self.M, model_arr.ravel()).reshape(data_arr.shape)
        return
    
    def adjoint(self, add, model, data):
        """m = A' * d"""
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        data_arr = data.getNdArray()
        model_arr = model.getNdArray()
        model_arr += np.matmul(self.M.T.conj(), data_arr.ravel()).reshape(model_arr.shape)
        return
    
    def getNdArray(self):
        return np.array(self.M)



# Module containing useful functions to interact with the Dask module
import atexit
import random
import socket
import subprocess
import os
import time
import json

DEVNULL = open(os.devnull, 'wb')
import dask.distributed as daskD
from dask_jobqueue import PBSCluster, LSFCluster, SLURMCluster
from dask_kubernetes import KubeCluster, make_pod_spec


def get_tcp_info(filename):
    """Function to obtain scheduler tcp information"""
    tcp_info = None
    with open(filename) as json_file:
        data = json.load(json_file)
    if "address" in data:
        tcp_info = data["address"]
    return tcp_info


def create_hostnames(machine_names, Nworkers):
    """
    Function to create hostnames variables (i.e., list of ip addresses)
    from machine names and number of wokers per machine
    """

    ip_adds = []
    for host in machine_names:
        ip_adds.append(socket.gethostbyname(host))

    if len(Nworkers) != len(ip_adds):
        raise ValueError("Lenght of number of workers (%s) not consistent with number of machines available (%s)"
                         % (len(Nworkers), len(ip_adds)))

    hostnames = []
    for idx, ip in enumerate(ip_adds):
        hostnames += [ip] * Nworkers[idx]
    return hostnames


def client_startup(cluster, n_jobs, total_workers):
    """
    Function to start a client
    """
    if n_jobs <= 0:
        raise ValueError("n_jobs must equal or greater than 1!")
    if isinstance(cluster, daskD.LocalCluster) or isinstance(cluster, KubeCluster):
        cluster.scale(n_jobs)
    else:
        cluster.scale(jobs=n_jobs)
    # Creating dask Client
    client = daskD.Client(cluster)
    workers = 0
    t0 = time.time()
    while workers < total_workers:
        workers = len(client.get_worker_logs().keys())
        # If the number of workers is not reached in 5 minutes raise exception
        if time.time() - t0 > 300.0:
            raise SystemError(
                "Dask could not start the requested workers within 5 minutes! Try different n_jobs.")
    WorkerIds = list(client.get_worker_logs().keys())
    return client, WorkerIds


class DaskClient:
    """
    Class useful to construct a Dask Client to be used with Dask vectors and operators
    """

    def __init__(self, **kwargs):
        """
    Constructor for obtaining a client to be used when Dask is necessary
    1) Cluster with shared file system and ssh capability:
    :param hostnames : - list; list of strings containing the host names or IP addresses of the machines that
    the user wants to use in their cluster/client (First hostname will be running the scheduler!) [None]
    :param scheduler_file_prefix : string; prefix to used to create dask scheduler-file.
    :param logging : - boolean; Logging scheduler and worker stdout to files within dask_logs folder [True]
    Must be a mounted path on all the machines. Necessary if hostnames are provided [$HOME/scheduler-]
    2) Local cluster:
    :param local_params : - dict; dictionary containing Local Cluster options (see help(LocalCluster) for help) [None]
    :param n_wrks: - int; number of workers to start [1]
    3) PBS cluster:
    :param pbs_params : - dict; dictionary containing PBS Cluster options (see help(PBSCluster) for help) [None]
    :param n_jobs : - int; number of jobs to be submitted to the cluster
    :param n_wrks: - int; number of workers per job [1]
    4) LSF cluster:
    :param lfs_params : - dict; dictionary containing LSF Cluster options (see help(LSFCluster) for help) [None]
    :param n_jobs : - int; number of jobs to be submitted to the cluster
    :param n_wrks: - int; number of workers per job [1]
    5) SLURM cluster:
    :param slurm_params : - dict; dictionary containing SLURM Cluster options (see help(SLURMCluster) for help) [None]
    :param n_jobs : - int; number of jobs to be submitted to the cluster
    :param n_wrks: - int; number of workers per job [1]
    6) Kubernetes cluster:
    :param kube_params : - dict; dictonary containing KubeCluster options
     (see help(KubeCluster) and help(make_pod_spec) for help) [None]
    :param n_wrks: - int; number of workers to scale the cluster
    Note that by default the Kubernetes pods are created using the Docker image 'ettore88/occamypy:devel'. To change
    the image to be use, provide the item image within the kube_params dictionary.
    """
        hostnames = kwargs.get("hostnames", None)
        local_params = kwargs.get("local_params", None)
        pbs_params = kwargs.get("pbs_params", None)
        lsf_params = kwargs.get("lsf_params", None)
        slurm_params = kwargs.get("slurm_params", None)
        kube_params = kwargs.get("kube_params", None)
        logging = kwargs.get("logging", True)
        ClusterInit = None
        cluster_params = None
        if local_params:
            cluster_params = local_params
            ClusterInit = daskD.LocalCluster
        elif pbs_params:
            cluster_params = pbs_params
            ClusterInit = PBSCluster
        elif lsf_params:
            cluster_params = lsf_params
            ClusterInit = LSFCluster
        elif slurm_params:
            cluster_params = slurm_params
            ClusterInit = SLURMCluster
        # Checking interface to be used
        if hostnames:
            if not isinstance(hostnames, list):
                raise ValueError("User must provide a list with host names")
            scheduler_file_prefix = kwargs.get("scheduler_file_prefix", os.path.expanduser("~") + "/scheduler-")
            # Random port number
            self.port = ''.join(["1"] + [str(random.randint(0, 9)) for _ in range(3)])
            # Creating logging interface
            stdout_scheduler = DEVNULL
            stdout_workers = [DEVNULL] * len(hostnames)
            if logging:
                # Creating logging folder
                try:
                    os.mkdir("dask_logs")
                except OSError:
                    pass
                stdout_scheduler = open("dask_logs/dask-scheduler.log", "w")
                stdout_workers = [open("dask_logs/dask-worker-%s.log" % (ii + 1), "w") for ii in range(len(hostnames))]
            # Starting scheduler
            scheduler_file = "%s%s" % (scheduler_file_prefix, self.port) + ".json"
            cmd = ["ssh"] + [hostnames[0]] + \
                  ["dask-scheduler"] + ["--scheduler-file"] + [scheduler_file] + \
                  ["--port"] + [self.port]
            self.scheduler_proc = subprocess.Popen(cmd, stdout=stdout_scheduler, stderr=subprocess.STDOUT)
            # Checking if scheduler has started and getting tpc information
            t0 = time.time()
            while True:
                if os.path.isfile(scheduler_file):
                    if get_tcp_info(scheduler_file): break
                # If the dask scheduler is not started in 5 minutes raise exception
                if time.time() - t0 > 300.0:
                    raise SystemError("Dask could not start scheduler! Try different first host name.")
            # Creating dask Client
            self.client = daskD.Client(scheduler_file=scheduler_file)
            # Starting workers on all the other hosts
            self.worker_procs = []
            worker_ips = []
            for ii, hostname in enumerate(hostnames):
                cmd = ["ssh"] + [hostname] + ["dask-worker"] + ["--scheduler-file"] + [scheduler_file]
                # Starting worker
                self.worker_procs.append(subprocess.Popen(cmd, stdout=stdout_workers[ii], stderr=subprocess.STDOUT))
                # Obtaining IP address of host for the started worker (necessary to resort workers)
                worker_ips.append(
                    subprocess.check_output(
                        ["ssh"] + [hostname] + ["hostname -I"] + ["| awk '{print $1}'"]).rstrip().decode("utf-8"))
            # Waiting until all the requested workers are up and running
            workers = 0
            requested = len(hostnames)
            t0 = time.time()
            while workers < requested:
                workers = len(self.client.get_worker_logs().keys())
                # If the number of workers is not reached in 5 minutes raise exception
                if time.time() - t0 > 300.0:
                    raise SystemError(
                        "Dask could not start the requested workers within 5 minutes! Try different hostnames.")
            # Resorting worker IDs according to user-provided list
            self.WorkerIds = []
            wrkIds = list(self.client.get_worker_logs().keys())  # Unsorted workers ids
            wrk_ips = [idw.split(":")[1][2:] for idw in wrkIds]  # Unsorted ip addresses
            for ip in worker_ips:
                idx = wrk_ips.index(ip)
                self.WorkerIds.append(wrkIds[idx])
                wrkIds.pop(idx)
                wrk_ips.pop(idx)
        elif kube_params:
            n_wrks = kwargs.get("n_wrks")
            if "image" not in kube_params:
                kube_params.update({"image": 'ettore88/occamypy:devel'})
            pod_spec = make_pod_spec(**kube_params)
            self.cluster = KubeCluster(pod_spec, deploy_mode="remote")
            self.client, self.WorkerIds = client_startup(self.cluster, n_wrks, n_wrks)
        elif ClusterInit:
            n_wrks = kwargs.get("n_wrks", 1)
            if n_wrks <= 0:
                raise ValueError("n_wrks must equal or greater than 1!")
            if "local_params" in kwargs:
                # Starting local cluster
                n_jobs = n_wrks
                n_wrks = 1
            else:
                # Starting scheduler-based clusters
                n_jobs = kwargs.get("n_jobs")
                if n_jobs <= 0:
                    raise ValueError("n_jobs must equal or greater than 1!")
                cluster_params.update({"processes": n_wrks})
                if n_wrks > 1:
                    # forcing nanny to be true (otherwise, dask-worker command will fail)
                    cluster_params.update({"nanny": True})
            self.cluster = ClusterInit(**cluster_params)
            self.client, self.WorkerIds = client_startup(self.cluster, n_jobs, n_jobs * n_wrks)
        else:
            raise ValueError("Either hostnames or local_params or pbs/lsf/slurm_params or kube_params must be "
                             "provided!")
        # Closing dask processes
        atexit.register(self.client.shutdown)

    def getClient(self):
        """
    Accessor for obtaining the client object
    """
        return self.client

    def getWorkerIds(self):
        """
    Accessor for obtaining the worker IDs
    """
        return self.WorkerIds

    def getNworkers(self):
        """
    Accessor for obtaining the number of workers
    """
        return len(self.getWorkerIds())



import time
from timeit import default_timer as timer
import numpy as np

from occamypy import problem as P


class Stopper:
    """Stopper parent object"""

    # Default class methods/functions
    def __init__(self):
        """Default class constructor for Stopper"""
        return

    def __del__(self):
        """Default destructor"""
        return

    def reset(self):
        """Function to reset stopper variables"""
        raise NotImplementedError("Implement reset stopper in the derived class.")

    def run(self, problem):
        """Dummy stopper running method"""
        raise NotImplementedError("Implement run stopper in the derived class.")


class BasicStopper(Stopper):
    """Basic Stopper with different options"""

    def __init__(self, niter=0, maxfevals=0, maxhours=0.0, tolr=1.0e-32, tolg=1.0e-32, tolg_proj=None, tolobj=None, tolobjrel=None,
                 toleta=None, tolobjchng=None, logger=None):
        """
        Constructor for Basic Stopper:
        niter    	 = [0] - integer; Number of iterations to run (must be greater than 0 to be checked)
        maxfevals    = [0] - integer; Maximum number of function evaluations (must be greater than 0 to be checked)
        maxhours     = [0.0] - float; Maximum total running time in hours (must be greater than 0.0 to be checked)
        tolr     	 = [1.0e-32] - float; Tolerance on residual norm
        tolg     	 = [1.0e-32] - float; Tolerance on gradient norm (Note: ignored for symmetric system)
        tolg_proj    = [None] - float; Tolerance on the projected-gradient infinity norm
        tolobj     	 = [None] - float; Tolerance on objective function value (Not relative value compared to initial one)
        tolobjrel    = [None] - float; Tolerance on relative objective function value (Should range between 0 and 1)
        toleta       = [None] - float; Tolerance on |Am - b|/|b| (Not supported for regularized problems)
        tolobjchng   = [None] - float; Tolerance on the change of the relative objective function value (phi(m_i) - phi(m_i-1)/ phi(m_0)) (Note that, the stopper averages 3 points by default, modify the internal variable 'ave_pts' to change the number of points)
        """
        # Criteria to evaluate whether or not to stop the solver
        super(BasicStopper, self).__init__()
        self.niter = niter
        self.zfill = int(np.floor(np.log10(self.niter)) + 1)  # number of digits for printing the iteration number
        self.maxfevals = maxfevals
        self.maxhours = maxhours
        self.tolr = tolr
        self.tolg = tolg
        self.tolg_proj = tolg_proj
        self.tolobj = tolobj
        self.tolobjrel = tolobjrel
        self.toleta = toleta
        # Tolerance on the change of the relative objective function value
        self.tolobjchng = tolobjchng
        # number of points to average plus the oldest one to compute objetive the change
        self.ave_pts = 3
        self.obj_pts = list()
        # Logger to write to file stopper information
        self.logger = logger
        # Starting timer
        self.__start = timer()
        return

    def reset(self):
        """Function to reset stopper variables"""
        # Restarting timer
        self.__start = timer()
        self.obj_pts = list()
        return

    # Beware stopper is going to change the gradient/obj/res files
    def run(self, problem, niter, initial_obj_value=None, verbose=False):
        if not isinstance(problem, P.Problem):
            raise TypeError("Input variable is not a Problem object")
        # Variable to impose stopping to solver
        stop = False
        # Taking time run so far (hours)
        elapsed_time = (timer() - self.__start) / 3600.0
        secs = elapsed_time * 3600.0
        # Printing elapsed time in hours, minutes, seconds
        hours = secs // 3600
        mins = (secs % 3600) // 60
        secs = (secs % 60)
        # Printing time stamp to log file if provided
        msg = "Elapsed time: %d hours, %d minutes, %d seconds\n" % (hours, mins, secs) + \
              "Current date & time: %s" % time.strftime("%c")
        res_norm = problem.get_rnorm(problem.model)
        try:
            grad_norm = problem.get_gnorm(problem.model)
        except NotImplementedError:
            grad_norm = None
        obj = problem.get_obj(problem.model)
        if self.logger:
            self.logger.addToLog(msg)
        # Stop by number of iterations
        if 0 < self.niter <= niter:
            stop = True
            msg = "Terminate: maximum number of iterations reached\n"
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            return stop
        if 0 < self.maxfevals <= problem.get_fevals():
            stop = True
            msg = "Terminate: maximum number of evaluations\n"
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            return stop
        if 0. < self.maxhours <= elapsed_time:
            stop = True
            msg = "Terminate: maximum number hours reached %s\n" % elapsed_time
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            return stop
        if res_norm < self.tolr:
            stop = True
            msg = "Terminate: minimum residual tolerance reached %s\n" % res_norm
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            return stop
        if grad_norm is not None and grad_norm < self.tolg:
            stop = True
            msg = "Terminate: minimum gradient tolerance reached %s\n" % grad_norm
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            return stop
        if grad_norm is not None and self.tolg_proj is not None:
            # Get the inf-norm of the projected gradient
            # Equation (6.1), Page 17. (L-BFGS-B)
            proj_grad = problem.model.clone()
            proj_grad.scaleAdd(problem.get_grad(problem.model), 1.0, -1.0)
            if "bounds" in dir(problem):
                problem.bounds.apply(proj_grad)
            proj_grad.scaleAdd(problem.model, 1.0, -1.0)
            grad_norm_proj = proj_grad.abs().max()
            del proj_grad
            if grad_norm_proj < self.tolg_proj:
                stop = True
                msg = "Terminate: tolerance of the projected gradient reached %s\n" % grad_norm_proj
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
                return stop
        if self.tolobj is not None:
            if obj < self.tolobj:
                stop = True
                msg = "Terminate: objective function value tolerance of %s reached, objective function value %s\n"\
                      % (self.tolobj, obj)
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
                return stop
        if self.tolobjrel is not None and initial_obj_value is not None and initial_obj_value != 0.:
            if obj / initial_obj_value < self.tolobjrel:
                stop = True
                msg = "Terminate: relative objective function value tolerance of %s reached, relative objective " \
                      "function value %s\n" % (self.tolobjrel, obj / initial_obj_value)
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
                return stop
        if self.toleta is not None:
            data_norm = problem.data.norm()
            if res_norm < self.toleta * data_norm:
                stop = True
                msg = "Terminate: eta tolerance (i.e., |Am - b|/|b|) of %s reached, eta value %s"\
                      % (self.toleta, res_norm / data_norm)
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
                return stop
        if self.tolobjchng is not None and initial_obj_value is not None and initial_obj_value != 0.0:
            # Saving current scaled objective function value
            self.obj_pts.append(obj / initial_obj_value)
            if niter >= self.ave_pts:
                # Appending initial objective function value
                objtmp = [1.0] + self.obj_pts
                objchng = np.abs(np.sum(np.diff(objtmp[-self.ave_pts - 1:])) / self.ave_pts)
                if objchng < self.tolobjchng:
                    stop = True
                    msg = "Terminate: tolerance on the change of the relative objective function value (i.e., " \
                          "phi(m_i) - phi(m_i-1)/ phi(m_0)) of %s reached (computed using %s points), change value " \
                          "%s" % (self.tolobjchng, self.ave_pts, objchng)
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog(msg)
                    return stop
        return stop


class SamplingStopper(Stopper):
    """Sampling Stopper with different options"""

    def __init__(self, nsamples, maxhours=0.0, logger=None):
        """
        Constructor for Sampling Stopper:
        nsamples - integer; Number of samples to tests
        maxhours - float; Maximum total running time in hours (must be greater than 0.0 to be checked) [0.0]
        """
        # Criteria to evaluate whether or not to stop the solver
        super(SamplingStopper, self).__init__()
        self.nsamples = nsamples
        self.zfill = int(np.floor(np.log10(self.nsamples)) + 1)  # number of digits for printing the iteration number
        self.maxhours = maxhours
        # Logger to write to file stopper information
        self.logger = logger
        # Starting timer
        self.__start = timer()
        return

    def reset(self):
        """Function to reset stopper variables"""
        # Restarting timer
        self.__start = timer()
        return

    # Beware stopper is going to change the gradient/obj/res files
    def run(self, problem, nsamples, verbose=False):
        if not isinstance(problem, P.Problem):
            raise TypeError("Input variable is not a Problem object")
        # Variable to impose stopping to solver
        stop = False
        # Taking time run so far (hours)
        elapsed_time = (timer() - self.__start) / 3600.0
        secs = elapsed_time * 3600.0
        # Printing elapsed time in hours, minutes, seconds
        hours = secs // 3600
        mins = (secs % 3600) // 60
        secs = (secs % 60)
        # Printing time stamp to log file if provided
        msg = "Elapsed time: %d hours, %d minutes, %d seconds\n" % (hours, mins, secs) + \
              "Current date & time: %s" % time.strftime("%c")
        if self.logger:
            self.logger.addToLog(msg)
        # Stop by number of iterations
        if 0 < self.nsamples <= nsamples:
            stop = True
            msg = "Terminate: maximum number of requested samples reached\n"
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            return stop
        if 0. < self.maxhours <= elapsed_time:
            stop = True
            msg = "Terminate: maximum number hours reached %s\n" % elapsed_time
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            return stop
        return stop



from .base import *
from .stopper import *
from .linear import *
from .nonlinear import *
from .sparsity import *
from .stepper import *

__all__ = [
    "Solver",
    "BasicStopper",
    "SamplingStopper",
    "CG",
    "SD",
    "LSQR",
    "CGsym",
    "NLCG",
    "LBFGS",
    "LBFGSB",
    "TNewton",
    "MCMC",
    "ISTA",
    "ISTC",
    "SplitBregman",
    "Stepper",
    "CvSrchStep",
    "ParabolicStep",
    "ParabolicStepConst",
    "StrongWolfe"
]



import torch
from math import sqrt
from numpy import ndarray
import importlib
GPUtil_found = importlib.util.find_spec("GPUtil")
if GPUtil_found:
    from GPUtil import getGPUs, getFirstAvailable


from occamypy import Vector


class VectorTorch(Vector):
    """In-core python vector class"""
    
    def __init__(self, in_content, device=None):
        """
        VectorTorch constructor: arr = torch.Tensor
        :param in_content: numpy ndarray or torch.Tensor or InCore Vector
        :param device: int - GPU id (None for CPU, -1 for most available memory)
        """
        if isinstance(in_content, Vector):
            try:
                self.arr = torch.from_numpy(in_content.getNdArray()).contiguous()
                self.ax_info = None
            except:
                raise UserWarning("Torch cannot handle the input array type")
        elif isinstance(in_content, ndarray):  # Numpy array passed to constructor
            self.arr = torch.from_numpy(in_content).contiguous()
            self.ax_info = None
        elif isinstance(in_content, torch.Tensor):  # Tensor passed to constructor
            self.arr = in_content.contiguous()
            self.ax_info = None
        elif isinstance(in_content, tuple):  # Tuple size passed to constructor
            self.arr = torch.empty(in_content)
            self.ax_info = None
        else:  # Not supported type
            raise ValueError("ERROR! Input variable not currently supported!")

        self.setDevice(device)
        
        super(VectorTorch, self).__init__()

        self.shape = tuple(self.arr.shape)  # Number of elements per axis (tuple)
        self.ndim = self.arr.ndim           # Number of axes (integer)
        self.size = self.arr.numel()        # Total number of elements (integer)
    
    def _check_same_device(self, other):
        assert isinstance(self, VectorTorch)
        assert isinstance(other, VectorTorch)
        answer = self.device == other.device
        if not answer:
            raise Warning('The two vectors live in different devices: %s - %s' % (self.device, other.device))
        return answer
    
    @property
    def device(self):
        try:
            return self.arr.device
        except AttributeError:
            return None

    def setDevice(self, devID=0):
        if devID is not None:  # Move to GPU
            if devID == -1:
                devID = getFirstAvailable(order='memory')[0]
            try:
                self.arr = self.arr.to(devID)
            except AttributeError:
                self.arr = self.arr.to('cpu')
                raise UserWarning("The selected device is not available, switched to CPU.")
        else:  # move to CPU
            if self.device is torch.device('cpu'):
                pass
            else:
                self.arr = self.arr.to('cpu')

    def printDevice(self):
        if self.device is torch.device('cpu'):
            print('CPU')
        else:
            name = getGPUs()[self.device.index].name
            print('GPU %d - %s' % (self.device.index, name))
        
    def getNdArray(self):
        """Function to return Ndarray of the vector"""
        return self.arr
    
    def norm(self, N=2):
        """Function to compute vector N-norm"""
        return torch.linalg.norm(self.getNdArray().flatten(), ord=N).item()
    
    def zero(self):
        """Function to zero out a vector"""
        self.set(0)
        return self
    
    def max(self):
        """Function to obtain maximum value in the vector"""
        return self.getNdArray().max().item()
    
    def min(self):
        """Function to obtain minimum value in the vector"""
        return self.getNdArray().min().item()
    
    def set(self, val: float or int):
        """Function to set all values in the vector"""
        self.getNdArray().fill_(val)
        return self
    
    def scale(self, sc: float or int):
        """Function to scale a vector"""
        self.getNdArray()[:] *= sc
        return self
    
    def addbias(self, bias: float or int):
        self.getNdArray()[:] += bias
        return self
    
    def rand(self, snr: float or int = 1.):
        """Fill vector with random number (~U[1,-1]) with a given SNR"""
        rms = torch.sqrt(torch.mean(torch.square(self.getNdArray())))
        amp_noise = 1.0
        if rms != 0.:
            amp_noise = sqrt(3. / snr) * rms  # sqrt(3*Power_signal/SNR)
        self.getNdArray()[:].uniform_(-1, 1)
        self.scale(amp_noise)
        return self
    
    def randn(self, snr=1.):
        """Fill vector with random number (~N[0,-1]) with a given SNR"""
        rms = torch.sqrt(torch.mean(torch.square(self.getNdArray())))
        amp_noise = 1.0
        if rms != 0.:
            amp_noise = sqrt(3. / snr) * rms  # sqrt(3*Power_signal/SNR)
        self.getNdArray()[:].normal_(0, 1)
        self.scale(amp_noise)
        return self
    
    def clone(self):
        """Function to clone (deep copy) a vector from a vector or a Space"""
        # If self is a Space vector, it is empty and it has only the shape, size and ndim attributes
        if self.getNdArray().numel() == 0:  # this is the shape of tensor!
            vec_clone = VectorTorch(torch.zeros(self.shape).type(self.getNdArray().dtype))
        
        else:  # self is a data vector, just clone
            vec_clone = VectorTorch(self.getNdArray().clone())
        
        vec_clone.setDevice(self.device.index)
        return vec_clone
    
    def cloneSpace(self):
        """Function to clone vector space only (vector without actual vector array by using empty array of size 0)"""
        vec_space = VectorTorch(torch.empty(0).type(self.getNdArray().dtype))
        vec_space.setDevice(self.device.index)
        # Cloning space of input vector
        vec_space.ndim = self.ndim
        vec_space.shape = self.shape
        vec_space.size = self.size
        return vec_space
    
    def checkSame(self, other):
        """Function to check dimensionality of vectors"""
        return self.shape == other.shape
    
    def abs(self):
        self.getNdArray().abs_()
        return self
    
    def sign(self):
        self.getNdArray().sign_()
        return self
    
    def reciprocal(self):
        self.getNdArray().reciprocal_()
        return self
    
    def maximum(self, other):
        if isinstance(other, (int, float)):
            # create a vector filled with the scalar value
            other = self.clone().set(other)
        
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        self.getNdArray()[:] = torch.maximum(self.getNdArray(), other.getNdArray())
        return self
    
    def conj(self):
        self.getNdArray()[:] = torch.conj(self.getNdArray())
        return self
    
    def transpose(self):
        other = VectorTorch(self.getNdArray().T)
        other.getNdArray()[:] = other.getNdArray()[:].contiguous()
        return other
    
    def pow(self, power):
        """Compute element-wise power of the vector"""
        self.getNdArray()[:].pow_(power)
        return self
    
    def real(self):
        """Return the real part of the vector"""
        self.getNdArray()[:] = torch.real(self.getNdArray())
        return self
    
    def imag(self, ):
        """Return the imaginary part of the vector"""
        self.getNdArray()[:] = torch.imag(self.getNdArray())
        return self
    
    def copy(self, other):
        """Function to copy vector from input vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorTorch):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        # Element-wise copy of the input array
        self.getNdArray()[:].copy_(other.getNdArray())
        return self
    
    def scaleAdd(self, other, sc1=1.0, sc2=1.0):
        """Function to scale a vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorTorch):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        # Performing scaling and addition
        self.scale(sc1)
        self.getNdArray()[:] += sc2 * other.getNdArray()
        return self
    
    def dot(self, other):
        """Function to compute dot product between two vectors"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorTorch):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking size (must have same number of elements)
        if self.size != other.size:
            raise ValueError("Vector size mismatching: self = %d; other = %d" % (self.size, other.size))
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        return torch.vdot(self.getNdArray().flatten(), other.getNdArray().flatten())
    
    def multiply(self, other):
        """Function to multiply element-wise two vectors"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorTorch):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking size (must have same number of elements)
        if self.size != other.size:
            raise ValueError("Vector size mismatching: self = %s; other = %s" % (self.size, other.size))
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        # Performing element-wise multiplication
        self.getNdArray()[:].multiply_(other.getNdArray())
        return self
    
    def isDifferent(self, other):
        """Function to check if two vectors are identical using built-in hash function"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorTorch):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        isDiff = not torch.equal(self.getNdArray(), other.getNdArray())
        return isDiff
    
    def clipVector(self, low, high):
        """Function to bound vector values based on input vectors low and high"""
        if not isinstance(low, VectorTorch):
            raise TypeError("Provided input low vector not a %s!" % self.whoami)
        if not isinstance(high, VectorTorch):
            raise TypeError("Provided input high vector not a %s!" % self.whoami)
        self.getNdArray()[:] = torch.minimum(torch.maximum(low.getNdArray(), self.getNdArray()), high.getNdArray())
        return self
    
    def plot(self):
        return self.getNdArray().detach().cpu().numpy()
